@@ -18,11 +18,13 @@ from __future__ import annotations
 
 import dataclasses
 import typing
-from typing import Any, Callable, ClassVar, Generic, Iterator, Optional, Tuple, Type, TypeVar, Union
+from typing import Any, Callable, ClassVar, Generic, Iterator, Optional, Set, Tuple, Type, TypeVar, Union
 
+from dataclass_array import field_utils
 from dataclass_array import shape_parsing
 from dataclass_array import type_parsing
-from dataclass_array.typing import Axes, DcOrArray, DcOrArrayT, DTypeArg, Shape  # pylint: disable=g-multiple-import
+from dataclass_array.typing import Array
+from dataclass_array.typing import Axes, DcOrArray, DcOrArrayT, DTypeArg, DynamicShape, Shape  # pylint: disable=g-multiple-import
 from dataclass_array.utils import np_utils
 from dataclass_array.utils import py_utils
 import einops
@@ -30,10 +32,9 @@ from etils import array_types
 from etils import edc
 from etils import enp
 from etils import epy
-from etils.array_types import Array
 import numpy as np
-from typing_extensions import Literal, TypeAlias  # pylint: disable=g-multiple-import
-
+import typing_extensions
+from typing_extensions import Annotated, Literal, TypeAlias  # pylint: disable=g-multiple-import
 
 lazy = enp.lazy
 
@@ -53,19 +54,76 @@ _METADATA_KEY = 'dca_field'
 class DataclassParams:
   """Params controlling the DataclassArray behavior.
 
-  Saved in `cls.__dca_params__`.
+  Set by `@dca.dataclass_array`. Saved in `cls.__dca_params__`.
 
   Attributes:
-    broadcast: If `False`, disable input broadcasting
-    cast_dtype: If `False`, do not auto-cast inputs `dtype`
-    cast_list: If `False`, do not auto-cast lists to `xnp.ndarray`
+    broadcast: If `True`, enable input broadcasting
+    cast_dtype: If `True`, auto-cast inputs `dtype`
+    cast_list: If `True`, auto-cast lists to `xnp.ndarray`
   """
-  broadcast: bool = True
-  cast_dtype: bool = True
+
+  # If modifying this, make sure to modify `@dataclass_array` too!
+  broadcast: bool = False
+  cast_dtype: bool = False
   cast_list: bool = True
 
 
-class DataclassArray:
+def dataclass_array(
+    *,
+    # If modifying this, make sure to modify `DataclassParams` too!
+    broadcast: bool = False,
+    cast_dtype: bool = False,
+    cast_list: bool = True,
+) -> Callable[[type[_DcT]], type[_DcT]]:
+  """Optional decorator to customize `dca.DataclassArray` params.
+
+  Usage:
+
+  ```python
+  @dca.dataclass_array()
+  @dataclasses.dataclass(frozen=True)
+  class MyDataclass(dca.DataclassArray):
+    ...
+  ```
+
+  This decorator has to be added in addition of inheriting from
+  `dca.DataclassArray`.
+
+  Args:
+    broadcast: If `True`, enable input broadcasting
+    cast_dtype: If `True`, auto-cast inputs `dtype`
+    cast_list: If `True`, auto-cast lists to `xnp.ndarray`
+
+  Returns:
+    decorator: The decorator which will apply the options to the dataclass
+  """
+
+  def decorator(cls):
+    if not issubclass(cls, DataclassArray):
+      raise TypeError(
+          '`@dca.dataclass_array` can only be applied on `dca.DataclassArray`. '
+          f'Got: {cls}'
+      )
+    cls.__dca_params__ = DataclassParams(
+        broadcast=broadcast,
+        cast_dtype=cast_dtype,
+        cast_list=cast_list,
+    )
+    return cls
+
+  return decorator
+
+
+class MetaDataclassArray(type):
+  """DataclassArray metaclass."""
+
+  # TODO(b/204422756): We cannot use `__class_getitem__` due to b/204422756
+  def __getitem__(cls, spec):
+    # Not clear how this would interact if cls is also a `Generic`
+    return Annotated[cls, field_utils.ShapeAnnotation(spec)]
+
+
+class DataclassArray(metaclass=MetaDataclassArray):
   """Dataclass which behaves like an array.
 
   Usage:
@@ -118,10 +176,16 @@ class DataclassArray:
   field annotated with `field: np.ndarray` or similar).
 
   """
+
   # Child class inherit the default params by default, but can also
   # overwrite them.
   __dca_params__: ClassVar[DataclassParams] = DataclassParams()
 
+  # TODO(epot): Could be removed with py3.10 and using `kw_only=True`
+  # Fields defined here will be forwarded with `.replace`
+  # TODO(py39): Replace Set -> set
+  __dca_non_init_fields__: ClassVar[Set[str]] = set()
+
   _shape: Shape
   _xnp: enp.NpModule
 
@@ -137,6 +201,11 @@ class DataclassArray:
     # convertions, we cache the type annotations here.
     cls._dca_fields_metadata: Optional[dict[str, _ArrayFieldMetadata]] = None
 
+    # Normalize the `cls.__dca_non_init_fields__`
+    # TODO(epot): Support inheritance if the parents also define
+    # `__dca_non_init_fields__` (fields should be merged from `.mro()`)
+    cls.__dca_non_init_fields__ = set(cls.__dca_non_init_fields__)
+
   def __post_init__(self) -> None:
     """Validate and normalize inputs."""
     cls = type(self)
@@ -144,7 +213,8 @@ class DataclassArray:
     # Make sure the dataclass was registered and frozen
     if not dataclasses.is_dataclass(cls) or not cls.__dataclass_params__.frozen:  # pytype: disable=attribute-error
       raise ValueError(
-          '`dca.DataclassArray` need to be @dataclasses.dataclass(frozen=True)')
+          '`dca.DataclassArray` need to be @dataclasses.dataclass(frozen=True)'
+      )
 
     # Register the tree_map here instead of `__init_subclass__` as `jax` may
     # not have been registered yet during import
@@ -155,7 +225,8 @@ class DataclassArray:
     if not self._all_array_fields:
       raise ValueError(
           f'{self.__class__.__qualname__} should have at least one '
-          '`dca.array_field`')
+          '`dca.array_field`'
+      )
 
     # Validate and normalize array fields
     # * Maybe cast (list, np) -> xnp
@@ -245,12 +316,10 @@ class DataclassArray:
 
   def broadcast_to(self: _DcT, shape: Shape) -> _DcT:
     """Broadcast the batch shape."""
-    # pyformat: disable
     return self._map_field(
         array_fn=lambda f: f.broadcast_to(shape),
         dc_fn=lambda f: f.broadcast_to(shape),
     )
-    # pyformat: enable
 
   def __getitem__(self: _DcT, indices: _IndicesArg) -> _DcT:
     """Slice indexing."""
@@ -278,7 +347,8 @@ class DataclassArray:
     """Length of the first array dimension."""
     if not self.shape:
       raise TypeError(
-          f'len() of unsized {self.__class__.__name__} (shape={self.shape})')
+          f'len() of unsized {self.__class__.__name__} (shape={self.shape})'
+      )
     return self.shape[0]
 
   def __bool__(self) -> Literal[True]:
@@ -320,7 +390,8 @@ class DataclassArray:
     if self.shape and not len(self):  # pylint: disable=g-explicit-length-test
       raise ValueError(
           f'The truth value of {self.__class__.__name__} when `len(x) == 0` '
-          'is ambigous. Use `len(x)` or `x is not None`.')
+          'is ambigous. Use `len(x)` or `x is not None`.'
+      )
     return True
 
   def map_field(
@@ -335,8 +406,32 @@ class DataclassArray:
 
   # ====== Dataclass/Conversion utils ======
 
-  # TODO(pytype): Could be removed once there's a way of annotating this.
-  replace = edc.dataclass_utils.replace
+  def replace(self: _DcT, **kwargs: Any) -> _DcT:
+    """Alias for `dataclasses.replace`."""
+    init_kwargs = {
+        k: v for k, v in kwargs.items() if k not in self.__dca_non_init_fields__
+    }
+    non_init_kwargs = {
+        k: v for k, v in kwargs.items() if k in self.__dca_non_init_fields__
+    }
+
+    # Create the new object
+    new_self = dataclasses.replace(self, **init_kwargs)
+
+    # Additionally forward the non-init kwargs
+    # `dataclasses.field(init=False) kwargs are required because `init=True`
+    # creates conflicts:
+    # * Inheritance fails with non-default argument 'K' follows default argument
+    # * Pytype complains too
+    # TODO(py310): Cleanup using `dataclass(kw_only)`
+    assert new_self is not self
+    for k in self.__dca_non_init_fields__:
+      if k in non_init_kwargs:
+        v = non_init_kwargs[k]
+      else:
+        v = getattr(self, k)
+      new_self._setattr(k, v)  # pylint: disable=protected-access
+    return new_self
 
   def as_np(self: _DcT) -> _DcT:
     """Returns the instance as containing `np.ndarray`."""
@@ -385,27 +480,34 @@ class DataclassArray:
     if cls._dca_fields_metadata is None:  # pylint: disable=protected-access
       # At this point, `ForwardRef` should have been resolved.
       try:
-        hints = typing.get_type_hints(cls)
+        hints = typing_extensions.get_type_hints(cls, include_extras=True)
       except Exception as e:  # pylint: disable=broad-except
-        epy.reraise(
-            e,
-            prefix=f'Could not infer typing annotation of {cls.__name__} '
-            f'defined in {cls.__module__}')
+        msg = (
+            f'Could not infer typing annotation of {cls.__qualname__} '
+            f'defined in {cls.__module__}:\n'
+        )
+        lines = [f' * {k}: {v!r}' for k, v in cls.__annotations__.items()]
+        lines = '\n'.join(lines)
+
+        epy.reraise(e, prefix=msg + lines + '\n')
 
       dca_fields_metadata = {
           f.name: _make_field_metadata(f, hints)
           for f in dataclasses.fields(self)
       }
-      cls._dca_fields_metadata = {  # Filter `None` values  # pylint: disable=protected-access
-          k: v for k, v in dca_fields_metadata.items() if v is not None
-      }
+      cls._dca_fields_metadata = (  # pylint: disable=protected-access
+          {  # Filter `None` values
+              k: v for k, v in dca_fields_metadata.items() if v is not None
+          }
+      )
 
     return {  # pylint: disable=g-complex-comprehension
         name: _ArrayField(
             name=name,
             host=self,
             **field_metadata.to_dict(),  # pylint: disable=not-a-mapping
-        ) for name, field_metadata in cls._dca_fields_metadata.items()  # pylint: disable=protected-access
+        )
+        for name, field_metadata in cls._dca_fields_metadata.items()  # pylint: disable=protected-access
     }
 
   @epy.cached_property
@@ -483,26 +585,27 @@ class DataclassArray:
     # Currently, we restrict broadcasting to either scalar or fixed length.
     # This is to avoid confusion broadcasting vs vectorization rules.
     # This restriction could be lifted if we encounter a use-case.
-    # pyformat: disable
     if (
         final_shape is None
         or len(shape_lengths) > 2
         or (len(shape_lengths) == 2 and 0 not in shape_lengths)
     ):
-      # pyformat: enable
       raise ValueError(
           f'Conflicting batch shapes: {shape_to_names}. '
           f'Currently {type(self).__qualname__}.__init__ broadcasting is '
           'restricted to scalar or dim=1 . '
-          'Please open an issue if you need more fine-grained broadcasting.')
+          'Please open an issue if you need more fine-grained broadcasting.'
+      )
 
     def _broadcast_field(f: _ArrayField) -> None:
       if f.host_shape == final_shape:  # Already broadcasted
         return
       elif not self.__dca_params__.broadcast:  # Broadcasing disabled
         raise ValueError(
-            f'{type(self).__qualname__} has `broadcast=False`. Cannot '
-            f'broadcast {f.name} from {f.full_shape} to {final_shape}')
+            f'{type(self).__qualname__} has `broadcast=False`. '
+            f'Cannot broadcast {f.name} from {f.full_shape} to {final_shape}. '
+            'To enable broadcast, use `@dca.dataclass_array(broadcast=True)`.'
+        )
       self._setattr(f.name, f.broadcast_to(final_shape))
 
     self._map_field(
@@ -579,7 +682,8 @@ class DataclassArray:
         zip(
             metadata.array_field_names,
             array_field_values,
-        ))
+        )
+    )
     init_fields = {}
     non_init_fields = {}
     fields = {f.name: f for f in dataclasses.fields(cls)}
@@ -592,11 +696,15 @@ class DataclassArray:
     self = cls(**array_field_kwargs, **init_fields)
     # Currently it's not clear how to handle non-init fields so raise an error
     if non_init_fields:
-      if set(non_init_fields) != {'fig_config'}:
+      if set(non_init_fields) - self.__dca_non_init_fields__:
         raise ValueError(
-            '`dca.DataclassArray` with init=False field not supported yet.')
+            '`dca.DataclassArray` field with init=False should be explicitly '
+            'specified in `__dca_non_init_fields__` for them to be '
+            'propagated by `tree_map`.'
+        )
       # TODO(py310): Delete once dataclass supports `kw_only=True`
-      self._setattr('fig_config', non_init_fields['fig_config'])  # pylint: disable=protected-access
+      for k, v in non_init_fields.items():
+        self._setattr(k, v)  # pylint: disable=protected-access
     return self
 
   def _setattr(self, name: str, value: Any) -> None:
@@ -609,7 +717,8 @@ class DataclassArray:
     if xnp is not self.xnp:
       raise ValueError(
           f'{self.__class__.__name__} is {self.xnp.__name__} but got input '
-          f'{xnp.__name__}. Please cast input first.')
+          f'{xnp.__name__}. Please cast input first.'
+      )
 
 
 def _infer_xnp(xnps: dict[enp.NpModule, list[str]]) -> enp.NpModule:
@@ -647,13 +756,15 @@ def _to_absolute_indices(indices: _Indices, *, shape: Shape) -> _Indices:
     raise IndexError("an index can only have a single ellipsis ('...')")
   valid_count = _count_not_none(indices)
   if valid_count > len(shape):
-    raise IndexError(f'too many indices for array. Batch shape is {shape}, but '
-                     f'rank-{valid_count} was provided.')
+    raise IndexError(
+        f'too many indices for array. Batch shape is {shape}, but '
+        f'rank-{valid_count} was provided.'
+    )
   if not ellipsis_count:
     return indices
   ellipsis_index = indices.index(Ellipsis)
   start_elems = indices[:ellipsis_index]
-  end_elems = indices[ellipsis_index + 1:]
+  end_elems = indices[ellipsis_index + 1 :]
   ellipsis_replacement = [slice(None)] * (len(shape) - valid_count)
   return (*start_elems, *ellipsis_replacement, *end_elems)
 
@@ -661,6 +772,7 @@ def _to_absolute_indices(indices: _Indices, *, shape: Shape) -> _Indices:
 @dataclasses.dataclass(frozen=True)
 class _TreeMetadata:
   """Metadata forwarded in ``."""
+
   array_field_names: list[str]
   non_array_field_kwargs: dict[str, Any]
 
@@ -706,7 +818,8 @@ class _ArrayFieldMetadata:
     dtype: Type of the array. Can be `array_types.dtypes.DType` or
       `dca.DataclassArray` for nested arrays.
   """
-  inner_shape_non_static: Shape
+
+  inner_shape_non_static: DynamicShape
   dtype: Union[array_types.dtypes.DType, type[DataclassArray]]
 
   def __post_init__(self):
@@ -738,8 +851,9 @@ class _ArrayField(_ArrayFieldMetadata, Generic[DcOrArrayT]):
     name: Instance of the attribute
     host: Dataclass instance who this field is attached too
   """
+
   name: str
-  host: DataclassArray
+  host: DataclassArray = dataclasses.field(repr=False)
 
   @property
   def qualname(self) -> str:
@@ -768,12 +882,13 @@ class _ArrayField(_ArrayFieldMetadata, Generic[DcOrArrayT]):
     """Returns the the static shape resolved for the current value."""
     if not self.inner_shape_non_static:
       return ()
-    static_shape = self.full_shape[-len(self.inner_shape_non_static):]
+    static_shape = self.full_shape[-len(self.inner_shape_non_static) :]
 
     def err_msg() -> ValueError:
       return ValueError(
           f'Shape do not match. Expected: {self.inner_shape_non_static}. '
-          f'Got {static_shape}')
+          f'Got {static_shape}'
+      )
 
     if len(static_shape) != len(self.inner_shape_non_static):
       raise err_msg()
@@ -795,9 +910,9 @@ class _ArrayField(_ArrayFieldMetadata, Generic[DcOrArrayT]):
       # In `jax/_src/api_util.py` for `flatten_axes`, jax set all values to a
       # dummy sentinel `object()` value.
       return True
-    elif (isinstance(self.value, DataclassArray) and
-          not self.value._array_fields  # pylint: disable=protected-access
-         ):
+    elif (
+        isinstance(self.value, DataclassArray) and not self.value._array_fields  # pylint: disable=protected-access
+    ):
       # Nested dataclass case (if all attributes are `None`, so no active
       # array fields)
       return True
@@ -809,14 +924,15 @@ class _ArrayField(_ArrayFieldMetadata, Generic[DcOrArrayT]):
     if not self.inner_shape_non_static:
       shape = self.full_shape
     else:
-      shape = self.full_shape[:-len(self.inner_shape_non_static)]
+      shape = self.full_shape[: -len(self.inner_shape_non_static)]
     return shape
 
   def assert_shape(self) -> None:
     if self.host_shape + self.inner_shape != self.full_shape:
       raise ValueError(
-          f'Shape should be '
-          f'{(py_utils.Ellipsis, *self.inner_shape)}. Got: {self.full_shape}')
+          'Shape should be '
+          f'{(py_utils.Ellipsis, *self.inner_shape)}. Got: {self.full_shape}'
+      )
 
   def broadcast_to(self, shape: Shape) -> DcOrArrayT:
     """Broadcast the host_shape."""
@@ -851,18 +967,18 @@ def _make_field_metadata(
 def _type_to_field_metadata(hint: TypeAlias) -> Optional[_ArrayFieldMetadata]:
   """Converts type hint to extract `inner_shape`, `dtype`."""
   array_type = type_parsing.get_array_type(hint)
-  if isinstance(array_type, type) and issubclass(array_type, DataclassArray):
-    # TODO(epot): Should support `ray: Ray[..., 3]` ?
-    return _ArrayFieldMetadata(inner_shape_non_static=(), dtype=array_type)
+
+  if isinstance(array_type, field_utils.DataclassWithShape):
+    dtype = array_type.cls
   elif isinstance(array_type, array_types.ArrayAliasMeta):
-    assert array_type is not None
-    try:
-      return _ArrayFieldMetadata(
-          inner_shape_non_static=shape_parsing.get_inner_shape(
-              array_type.shape),
-          dtype=array_type.dtype,
-      )
-    except Exception as e:  # pylint: disable=broad-except
-      epy.reraise(e, prefix=f'Invalid shape annotation {hint}.')
+    dtype = array_type.dtype
   else:  # Not a supported type: Static field
     return None
+
+  try:
+    return _ArrayFieldMetadata(
+        inner_shape_non_static=shape_parsing.get_inner_shape(array_type.shape),
+        dtype=dtype,
+    )
+  except Exception as e:  # pylint: disable=broad-except
+    epy.reraise(e, prefix=f'Invalid shape annotation {hint}.')
